@@ -1,17 +1,15 @@
+'use strict';
 var http = require('http');
 
 var concurrency = 30;
 var port = 12346;
 var n = 700;
-var bytes = 1024*5;
+var bytes = 1024 * 5;
 
 var requests = 0;
 var responses = 0;
 
-var body = '';
-for (var i = 0; i < bytes; i++) {
-  body += 'C';
-}
+var body = 'C'.repeat(bytes);
 
 var server = http.createServer(function(req, res) {
   res.writeHead(200, {
@@ -19,7 +17,7 @@ var server = http.createServer(function(req, res) {
     'Content-Length': body.length
   });
   res.end(body);
-})
+});
 
 server.listen(port, function() {
   var agent = new http.Agent();
