@@ -1015,11 +1015,7 @@ function leftPad(n, prefix, maxN) {
   const nchars = Math.max(2, String(maxN).length) + 1;
   const nspaces = nchars - s.length - 1;
 
-  for (var i = 0; i < nspaces; i++) {
-    prefix += ' ';
-  }
-
-  return prefix + s;
+  return prefix + ' '.repeat(nspaces) + s;
 }
 
 
@@ -1353,7 +1349,7 @@ Interface.prototype.setBreakpoint = function(script, line,
   }
 
   let req;
-  if (/\(\)$/.test(script)) {
+  if (script.endsWith('()')) {
     // setBreakpoint('functionname()');
     req = {
       type: 'function',
