@@ -87,6 +87,17 @@ function ClientRequest(options, cb) {
     }
     if (host && !this.getHeader('host') && setHost) {
       var hostHeader = host;
+      var posColon = -1;
+
+      // For the Host header, ensure that IPv6 addresses are enclosed
+      // in square brackets, as defined by URI formatting
+      // https://tools.ietf.org/html/rfc3986#section-3.2.2
+      if (-1 !== (posColon = hostHeader.indexOf(':')) &&
+          -1 !== (posColon = hostHeader.indexOf(':', posColon)) &&
+          '[' !== hostHeader[0]) {
+        hostHeader = `[${hostHeader}]`;
+      }
+
       if (port && +port !== defaultPort) {
         hostHeader += ':' + port;
       }
