@@ -89,8 +89,10 @@ exports.OutgoingMessage = OutgoingMessage;
 
 
 OutgoingMessage.prototype.setTimeout = function(msecs, callback) {
-  if (callback)
+
+  if (callback) {
     this.on('timeout', callback);
+  }
 
   if (!this.socket) {
     this.once('socket', function(socket) {
@@ -299,8 +301,14 @@ OutgoingMessage.prototype._storeHeader = function(firstLine, headers) {
 };
 
 function storeHeader(self, state, field, value) {
-  value = escapeHeaderValue(value);
-  state.messageHeader += field + ': ' + value + CRLF;
+  if (!common._checkIsHttpToken(field)) {
+    throw new TypeError(
+      'Header name must be a valid HTTP Token ["' + field + '"]');
+  }
+  if (common._checkInvalidHeaderChar(value) === true) {
+    throw new TypeError('The header content contains invalid characters');
+  }
+  state.messageHeader += field + ': ' + escapeHeaderValue(value) + CRLF;
 
   if (connectionExpression.test(field)) {
     state.sentConnectionHeader = true;
@@ -327,13 +335,18 @@ function storeHeader(self, state, field, value) {
 
 
 OutgoingMessage.prototype.setHeader = function(name, value) {
+  if (!common._checkIsHttpToken(name))
+    throw new TypeError(
+      'Header name must be a valid HTTP Token ["' + name + '"]');
   if (typeof name !== 'string')
     throw new TypeError('`name` should be a string in setHeader(name, value).');
   if (value === undefined)
     throw new Error('`value` required in setHeader("' + name + '", value).');
   if (this._header)
     throw new Error('Can\'t set headers after they are sent.');
-
+  if (common._checkInvalidHeaderChar(value) === true) {
+    throw new TypeError('The header content contains invalid characters');
+  }
   if (this._headers === null)
     this._headers = {};
 
@@ -476,8 +489,7 @@ function writeAfterEndNT(self, err, callback) {
 
 
 function connectionCorkNT(conn) {
-  if (conn)
-    conn.uncork();
+  conn.uncork();
 }
 
 
@@ -502,7 +514,13 @@ OutgoingMessage.prototype.addTrailers = function(headers) {
       field = key;
       value = headers[key];
     }
-
+    if (!common._checkIsHttpToken(field)) {
+      throw new TypeError(
+        'Trailer name must be a valid HTTP Token ["' + field + '"]');
+    }
+    if (common._checkInvalidHeaderChar(value) === true) {
+      throw new TypeError('The header content contains invalid characters');
+    }
     this._trailer += field + ': ' + escapeHeaderValue(value) + CRLF;
   }
 };
@@ -560,7 +578,7 @@ OutgoingMessage.prototype.end = function(data, encoding, callback) {
   var ret;
   if (data) {
     // Normal body write.
-    ret = this.write(data, encoding);
+    this.write(data, encoding);
   }
 
   if (this._hasBody && this.chunkedEncoding) {
