@@ -1,49 +1,40 @@
-// Copyright Joyent, Inc. and other Node contributors.
-//
-// // Emit `beforeExit` if the loop became alive either after emitting
-// event, or after running some callbacks.
-//
-// Permission is hereby granted, free of charge, to any person obtaining a
-// copy of this software and associated documentation files (the
-// "Software"), to deal in the Software without restriction, including
-// without limitation the rights to use, copy, modify, merge, publish,
-// distribute, sublicense, and/or sell copies of the Software, and to permit
-// persons to whom the Software is furnished to do so, subject to the
-// following conditions:
-//
-// The above copyright notice and this permission notice shall be included
-// in all copies or substantial portions of the Software.
-//
-// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
-// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
-// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
-// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
-// USE OR OTHER DEALINGS IN THE SOFTWARE.
+'use strict';
 
-var assert = require('assert');
-var crypto = require('crypto');
-var net = require('net');
-var tls = require('tls');
-var util = require('util');
-var listenerCount = require('events').listenerCount;
-var common = require('_tls_common');
-var constants = require('constants');
+const assert = require('assert');
+const crypto = require('crypto');
+const net = require('net');
+const tls = require('tls');
+const util = require('util');
+const common = require('_tls_common');
+const StreamWrap = require('_stream_wrap').StreamWrap;
+const Buffer = require('buffer').Buffer;
+const Duplex = require('stream').Duplex;
+const debug = util.debuglog('tls');
+const Timer = process.binding('timer_wrap').Timer;
+const tls_wrap = process.binding('tls_wrap');
+const TCP = process.binding('tcp_wrap').TCP;
+const Pipe = process.binding('pipe_wrap').Pipe;
+const defaultSessionIdContext = getDefaultSessionIdContext();
 
-var Timer = process.binding('timer_wrap').Timer;
-var tls_wrap = process.binding('tls_wrap');
-
-// Lazy load
-var tls_legacy;
-
-var debug = util.debuglog('tls');
+function getDefaultSessionIdContext() {
+  var defaultText = process.argv.join(' ');
+  /* SSL_MAX_SID_CTX_LENGTH is 128 bits */
+  if (process.config.variables.openssl_fips) {
+    return crypto.createHash('sha1')
+      .update(defaultText)
+      .digest('hex').slice(0, 32);
+  } else {
+    return crypto.createHash('md5')
+      .update(defaultText)
+      .digest('hex');
+  }
+}
 
 function onhandshakestart() {
   debug('onhandshakestart');
 
   var self = this;
-  var ssl = self.ssl;
+  var ssl = self._handle;
   var now = Timer.now();
 
   assert(now >= ssl.lastHandshakeTime);
@@ -62,7 +53,7 @@ function onhandshakestart() {
     // callback to destroy the connection right now, it would crash and burn.
     setImmediate(function() {
       var err = new Error('TLS session renegotiation attack detected.');
-      self._tlsError(err);
+      self._emitTLSError(err);
     });
   }
 }
@@ -85,10 +76,13 @@ function loadSession(self, hello, cb) {
     if (err)
       return cb(err);
 
+    if (!self._handle)
+      return cb(new Error('Socket is closed'));
+
     // NOTE: That we have disabled OpenSSL's internal session storage in
     // `node_crypto.cc` and hence its safe to rely on getting servername only
     // from clienthello or this place.
-    var ret = self.ssl.loadSession(session);
+    var ret = self._handle.loadSession(session);
 
     cb(null, ret);
   }
@@ -115,11 +109,14 @@ function loadSNI(self, servername, cb) {
     if (err)
       return cb(err);
 
+    if (!self._handle)
+      return cb(new Error('Socket is closed'));
+
     // TODO(indutny): eventually disallow raw `SecureContext`
     if (context)
-      self.ssl.sni_context = context.context || context;
+      self._handle.sni_context = context.context || context;
 
-    cb(null, self.ssl.sni_context);
+    cb(null, self._handle.sni_context);
   });
 }
 
@@ -133,7 +130,7 @@ function requestOCSP(self, hello, ctx, cb) {
   if (ctx.context)
     ctx = ctx.context;
 
-  if (listenerCount(self.server, 'OCSPRequest') === 0) {
+  if (self.server.listenerCount('OCSPRequest') === 0) {
     return cb(null);
   } else {
     self.server.emit('OCSPRequest',
@@ -151,8 +148,11 @@ function requestOCSP(self, hello, ctx, cb) {
     if (err)
       return cb(err);
 
+    if (!self._handle)
+      return cb(new Error('Socket is closed'));
+
     if (response)
-      self.ssl.setOCSPResponse(response);
+      self._handle.setOCSPResponse(response);
     cb(null);
   }
 }
@@ -165,29 +165,26 @@ function onclienthello(hello) {
     if (err)
       return self.destroy(err);
 
-    // Servername came from SSL session
-    // NOTE: TLS Session ticket doesn't include servername information
-    //
-    // Another note, From RFC3546:
-    //
-    //   If, on the other hand, the older
-    //   session is resumed, then the server MUST ignore extensions appearing
-    //   in the client hello, and send a server hello containing no
-    //   extensions; in this case the extension functionality negotiated
-    //   during the original session initiation is applied to the resumed
-    //   session.
-    //
-    // Therefore we should account session loading when dealing with servername
-    var servername = session && session.servername || hello.servername;
-    loadSNI(self, servername, function(err, ctx) {
+    self._handle.endParser();
+  });
+}
+
+
+function oncertcb(info) {
+  var self = this;
+  var servername = info.servername;
+
+  loadSNI(self, servername, function(err, ctx) {
+    if (err)
+      return self.destroy(err);
+    requestOCSP(self, info, ctx, function(err) {
       if (err)
         return self.destroy(err);
-      requestOCSP(self, hello, ctx, function(err) {
-        if (err)
-          return self.destroy(err);
 
-        self.ssl.endParser();
-      });
+      if (!self._handle)
+        return self.destroy(new Error('Socket is closed'));
+
+      self._handle.certCbDone();
     });
   });
 }
@@ -209,7 +206,10 @@ function onnewsession(key, session) {
       return;
     once = true;
 
-    self.ssl.newSessionDone();
+    if (!self._handle)
+      return self.destroy(new Error('Socket is closed'));
+
+    self._handle.newSessionDone();
 
     self._newSessionPending = false;
     if (self._securePending)
@@ -223,128 +223,219 @@ function onocspresponse(resp) {
   this.emit('OCSPResponse', resp);
 }
 
+function initRead(tls, wrapped) {
+  // If we were destroyed already don't bother reading
+  if (!tls._handle)
+    return;
+
+  // Socket already has some buffered data - emulate receiving it
+  if (wrapped && wrapped._readableState && wrapped._readableState.length) {
+    var buf;
+    while ((buf = wrapped.read()) !== null)
+      tls._handle.receive(buf);
+  }
+
+  tls.read(0);
+}
 
 /**
  * Provides a wrap of socket stream to do encrypted communication.
  */
 
 function TLSSocket(socket, options) {
-  // Disallow wrapping TLSSocket in TLSSocket
-  assert(!(socket instanceof TLSSocket));
-  var self = this;
-
-  net.Socket.call(this, {
-    handle: socket && socket._handle,
-    allowHalfOpen: socket && socket.allowHalfOpen,
-    readable: false,
-    writable: false
-  });
-
-  if (socket) {
-    this._parent = socket;
-    socket._destroy = function(exception) {
-      self._destroy(exception);
-    };
-    // To prevent assertion in afterConnect()
-    this._connecting = socket._connecting;
-  }
-
-  this._tlsOptions = options;
+  if (options === undefined)
+    this._tlsOptions = {};
+  else
+    this._tlsOptions = options;
   this._secureEstablished = false;
   this._securePending = false;
   this._newSessionPending = false;
   this._controlReleased = false;
   this._SNICallback = null;
-  this.ssl = null;
   this.servername = null;
   this.npnProtocol = null;
+  this.alpnProtocol = null;
   this.authorized = false;
   this.authorizationError = null;
 
+  // Wrap plain JS Stream into StreamWrap
+  var wrap;
+  if (!(socket instanceof net.Socket) && socket instanceof Duplex)
+    wrap = new StreamWrap(socket);
+  else if ((socket instanceof net.Socket) && !socket._handle)
+    wrap = new StreamWrap(socket);
+  else
+    wrap = socket;
+
   // Just a documented property to make secure sockets
   // distinguishable from regular ones.
   this.encrypted = true;
 
-  this.on('error', this._tlsError);
+  net.Socket.call(this, {
+    handle: this._wrapHandle(wrap),
+    allowHalfOpen: socket && socket.allowHalfOpen,
+    readable: false,
+    writable: false
+  });
 
-  if (!this._handle) {
-    this.once('connect', function() {
-      this._init(null);
-    });
-  } else {
-    this._init(socket);
-  }
+  // Proxy for API compatibility
+  this.ssl = this._handle;
+
+  this.on('error', this._emitTLSError);
+
+  this._init(socket, wrap);
 
   // Make sure to setup all required properties like: `_connecting` before
   // starting the flow of the data
   this.readable = true;
   this.writable = true;
-  this.read(0);
+
+  // Read on next tick so the caller has a chance to setup listeners
+  process.nextTick(initRead, this, socket);
 }
 util.inherits(TLSSocket, net.Socket);
 exports.TLSSocket = TLSSocket;
 
-TLSSocket.prototype._init = function(socket) {
-  assert(this._handle);
+var proxiedMethods = [
+  'ref', 'unref', 'open', 'bind', 'listen', 'connect', 'bind6',
+  'connect6', 'getsockname', 'getpeername', 'setNoDelay', 'setKeepAlive',
+  'setSimultaneousAccepts', 'setBlocking',
 
-  // lib/net.js expect this value to be non-zero if write hasn't been flushed
-  // immediately
-  // TODO(indutny): rewise this solution, it might be 1 before handshake and
-  // represent real writeQueueSize during regular writes.
-  this._handle.writeQueueSize = 1;
+  // PipeWrap
+  'setPendingInstances'
+];
+
+// Proxy HandleWrap, PipeWrap and TCPWrap methods
+proxiedMethods.forEach(function(name) {
+  tls_wrap.TLSWrap.prototype[name] = function methodProxy() {
+    if (this._parent[name])
+      return this._parent[name].apply(this._parent, arguments);
+  };
+});
+
+tls_wrap.TLSWrap.prototype.close = function closeProxy(cb) {
+  if (this.owner)
+    this.owner.ssl = null;
+
+  if (this._parentWrap && this._parentWrap._handle === this._parent) {
+    this._parentWrap.once('close', cb);
+    return this._parentWrap.destroy();
+  }
+  return this._parent.close(cb);
+};
+
+TLSSocket.prototype._wrapHandle = function(wrap) {
+  var res;
+  var handle;
+
+  if (wrap)
+    handle = wrap._handle;
 
-  var self = this;
   var options = this._tlsOptions;
+  if (!handle) {
+    handle = options.pipe ? new Pipe() : new TCP();
+    handle.owner = this;
+  }
 
   // Wrap socket's handle
   var context = options.secureContext ||
                 options.credentials ||
                 tls.createSecureContext();
-  this.ssl = tls_wrap.wrap(this._handle, context.context, options.isServer);
-  this.server = options.server || null;
+  res = tls_wrap.wrap(handle._externalStream,
+                      context.context,
+                      !!options.isServer);
+  res._parent = handle;
+  res._parentWrap = wrap;
+  res._secureContext = context;
+  res.reading = handle.reading;
+  Object.defineProperty(handle, 'reading', {
+    get: function readingGetter() {
+      return res.reading;
+    },
+    set: function readingSetter(value) {
+      res.reading = value;
+    }
+  });
+
+  this.on('close', function() {
+    // Make sure we are not doing it on OpenSSL's stack
+    setImmediate(destroySSL, this);
+    res = null;
+  });
+
+  return res;
+};
+
+function destroySSL(self) {
+  self._destroySSL();
+}
+
+TLSSocket.prototype._destroySSL = function _destroySSL() {
+  if (!this.ssl) return;
+  this.ssl.destroySSL();
+  if (this.ssl._secureContext.singleUse) {
+    this.ssl._secureContext.context.close();
+    this.ssl._secureContext.context = null;
+  }
+  this.ssl = null;
+};
+
+TLSSocket.prototype._init = function(socket, wrap) {
+  var self = this;
+  var options = this._tlsOptions;
+  var ssl = this._handle;
+
+  // lib/net.js expect this value to be non-zero if write hasn't been flushed
+  // immediately
+  // TODO(indutny): rewise this solution, it might be 1 before handshake and
+  // represent real writeQueueSize during regular writes.
+  ssl.writeQueueSize = 1;
+
+  this.server = options.server;
 
   // For clients, we will always have either a given ca list or be using
   // default one
-  var requestCert = !!options.requestCert || !options.isServer,
-      rejectUnauthorized = !!options.rejectUnauthorized;
+  const requestCert = !!options.requestCert || !options.isServer;
+  const rejectUnauthorized = !!options.rejectUnauthorized;
 
   this._requestCert = requestCert;
   this._rejectUnauthorized = rejectUnauthorized;
   if (requestCert || rejectUnauthorized)
-    this.ssl.setVerifyMode(requestCert, rejectUnauthorized);
+    ssl.setVerifyMode(requestCert, rejectUnauthorized);
 
   if (options.isServer) {
-    this.ssl.onhandshakestart = onhandshakestart.bind(this);
-    this.ssl.onhandshakedone = onhandshakedone.bind(this);
-    this.ssl.onclienthello = onclienthello.bind(this);
-    this.ssl.onnewsession = onnewsession.bind(this);
-    this.ssl.lastHandshakeTime = 0;
-    this.ssl.handshakes = 0;
+    ssl.onhandshakestart = () => onhandshakestart.call(this);
+    ssl.onhandshakedone = () => onhandshakedone.call(this);
+    ssl.onclienthello = (hello) => onclienthello.call(this, hello);
+    ssl.oncertcb = (info) => oncertcb.call(this, info);
+    ssl.onnewsession = (key, session) => onnewsession.call(this, key, session);
+    ssl.lastHandshakeTime = 0;
+    ssl.handshakes = 0;
 
-    if (this.server &&
-        (listenerCount(this.server, 'resumeSession') > 0 ||
-         listenerCount(this.server, 'newSession') > 0 ||
-         listenerCount(this.server, 'OCSPRequest') > 0)) {
-      this.ssl.enableSessionCallbacks();
+    if (this.server) {
+      if (this.server.listenerCount('resumeSession') > 0 ||
+          this.server.listenerCount('newSession') > 0) {
+        ssl.enableSessionCallbacks();
+      }
+      if (this.server.listenerCount('OCSPRequest') > 0)
+        ssl.enableCertCb();
     }
   } else {
-    this.ssl.onhandshakestart = function() {};
-    this.ssl.onhandshakedone = this._finishInit.bind(this);
-    this.ssl.onocspresponse = onocspresponse.bind(this);
+    ssl.onhandshakestart = function() {};
+    ssl.onhandshakedone = () => this._finishInit();
+    ssl.onocspresponse = (resp) => onocspresponse.call(this, resp);
 
     if (options.session)
-      this.ssl.setSession(options.session);
+      ssl.setSession(options.session);
   }
 
-  this.ssl.onerror = function(err) {
+  ssl.onerror = function(err) {
     if (self._writableState.errorEmitted)
       return;
-    self._writableState.errorEmitted = true;
 
     // Destroy socket if error happened before handshake's finish
-    if (!this._secureEstablished) {
-      self._tlsError(err);
-      self.destroy();
+    if (!self._secureEstablished) {
+      self.destroy(self._tlsError(err));
     } else if (options.isServer &&
                rejectUnauthorized &&
                /peer did not return a certificate/.test(err.message)) {
@@ -352,8 +443,10 @@ TLSSocket.prototype._init = function(socket) {
       self.destroy();
     } else {
       // Throw error
-      self._tlsError(err);
+      self._emitTLSError(err);
     }
+
+    self._writableState.errorEmitted = true;
   };
 
   // If custom SNICallback was given, or if
@@ -361,31 +454,55 @@ TLSSocket.prototype._init = function(socket) {
   // set `.onsniselect` callback.
   if (process.features.tls_sni &&
       options.isServer &&
+      options.SNICallback &&
       options.server &&
       (options.SNICallback !== SNICallback ||
        options.server._contexts.length)) {
     assert(typeof options.SNICallback === 'function');
     this._SNICallback = options.SNICallback;
-    this.ssl.enableHelloParser();
+    ssl.enableCertCb();
   }
 
   if (process.features.tls_npn && options.NPNProtocols)
-    this.ssl.setNPNProtocols(options.NPNProtocols);
+    ssl.setNPNProtocols(options.NPNProtocols);
+
+  if (process.features.tls_alpn && options.ALPNProtocols) {
+    // keep reference in secureContext not to be GC-ed
+    ssl._secureContext.alpnBuffer = options.ALPNProtocols;
+    ssl.setALPNProtocols(ssl._secureContext.alpnBuffer);
+  }
 
   if (options.handshakeTimeout > 0)
     this.setTimeout(options.handshakeTimeout, this._handleTimeout);
 
-  // Socket already has some buffered data - emulate receiving it
-  if (socket && socket._readableState.length) {
-    var buf;
-    while ((buf = socket.read()) !== null)
-      this.ssl.receive(buf);
+  if (socket instanceof net.Socket) {
+    this._parent = socket;
+
+    // To prevent assertion in afterConnect() and properly kick off readStart
+    this._connecting = socket._connecting || !socket._handle;
+    socket.once('connect', function() {
+      self._connecting = false;
+      self.emit('connect');
+    });
+  }
+
+  // Assume `tls.connect()`
+  if (wrap) {
+    wrap.on('error', function(err) {
+      self._emitTLSError(err);
+    });
+  } else {
+    assert(!socket);
+    this._connecting = true;
   }
 };
 
 TLSSocket.prototype.renegotiate = function(options, callback) {
-  var requestCert = this._requestCert,
-      rejectUnauthorized = this._rejectUnauthorized;
+  var requestCert = this._requestCert;
+  var rejectUnauthorized = this._rejectUnauthorized;
+
+  if (this.destroyed)
+    return;
 
   if (typeof options.requestCert !== 'undefined')
     requestCert = !!options.requestCert;
@@ -394,15 +511,13 @@ TLSSocket.prototype.renegotiate = function(options, callback) {
 
   if (requestCert !== this._requestCert ||
       rejectUnauthorized !== this._rejectUnauthorized) {
-    this.ssl.setVerifyMode(requestCert, rejectUnauthorized);
+    this._handle.setVerifyMode(requestCert, rejectUnauthorized);
     this._requestCert = requestCert;
     this._rejectUnauthorized = rejectUnauthorized;
   }
-  if (!this.ssl.renegotiate()) {
+  if (!this._handle.renegotiate()) {
     if (callback) {
-      process.nextTick(function() {
-        callback(new Error('Failed to renegotiate'));
-      });
+      process.nextTick(callback, new Error('Failed to renegotiate'));
     }
     return false;
   }
@@ -420,28 +535,35 @@ TLSSocket.prototype.renegotiate = function(options, callback) {
 };
 
 TLSSocket.prototype.setMaxSendFragment = function setMaxSendFragment(size) {
-  return this.ssl.setMaxSendFragment(size) == 1;
+  return this._handle.setMaxSendFragment(size) == 1;
 };
 
 TLSSocket.prototype.getTLSTicket = function getTLSTicket() {
-  return this.ssl.getTLSTicket();
+  return this._handle.getTLSTicket();
 };
 
 TLSSocket.prototype._handleTimeout = function() {
-  this._tlsError(new Error('TLS handshake timeout'));
+  this._emitTLSError(new Error('TLS handshake timeout'));
+};
+
+TLSSocket.prototype._emitTLSError = function(err) {
+  var e = this._tlsError(err);
+  if (e)
+    this.emit('error', e);
 };
 
 TLSSocket.prototype._tlsError = function(err) {
   this.emit('_tlsError', err);
   if (this._controlReleased)
-    this.emit('error', err);
+    return err;
+  return null;
 };
 
 TLSSocket.prototype._releaseControl = function() {
   if (this._controlReleased)
     return false;
   this._controlReleased = true;
-  this.removeListener('error', this._tlsError);
+  this.removeListener('error', this._emitTLSError);
   return true;
 };
 
@@ -453,11 +575,15 @@ TLSSocket.prototype._finishInit = function() {
   }
 
   if (process.features.tls_npn) {
-    this.npnProtocol = this.ssl.getNegotiatedProtocol();
+    this.npnProtocol = this._handle.getNegotiatedProtocol();
+  }
+
+  if (process.features.tls_alpn) {
+    this.alpnProtocol = this.ssl.getALPNNegotiatedProtocol();
   }
 
   if (process.features.tls_sni && this._tlsOptions.isServer) {
-    this.servername = this.ssl.getServername();
+    this.servername = this._handle.getServername();
   }
 
   debug('secure established');
@@ -468,54 +594,80 @@ TLSSocket.prototype._finishInit = function() {
 };
 
 TLSSocket.prototype._start = function() {
+  if (this._connecting) {
+    this.once('connect', function() {
+      this._start();
+    });
+    return;
+  }
+
+  // Socket was destroyed before the connection was established
+  if (!this._handle)
+    return;
+
+  debug('start');
   if (this._tlsOptions.requestOCSP)
-    this.ssl.requestOCSP();
-  this.ssl.start();
+    this._handle.requestOCSP();
+  this._handle.start();
 };
 
 TLSSocket.prototype.setServername = function(name) {
-  this.ssl.setServername(name);
+  this._handle.setServername(name);
 };
 
 TLSSocket.prototype.setSession = function(session) {
-  if (util.isString(session))
+  if (typeof session === 'string')
     session = new Buffer(session, 'binary');
-  this.ssl.setSession(session);
+  this._handle.setSession(session);
 };
 
 TLSSocket.prototype.getPeerCertificate = function(detailed) {
-  if (this.ssl) {
+  if (this._handle) {
     return common.translatePeerCertificate(
-        this.ssl.getPeerCertificate(detailed));
+        this._handle.getPeerCertificate(detailed));
   }
 
   return null;
 };
 
 TLSSocket.prototype.getSession = function() {
-  if (this.ssl) {
-    return this.ssl.getSession();
+  if (this._handle) {
+    return this._handle.getSession();
   }
 
   return null;
 };
 
 TLSSocket.prototype.isSessionReused = function() {
-  if (this.ssl) {
-    return this.ssl.isSessionReused();
+  if (this._handle) {
+    return this._handle.isSessionReused();
   }
 
   return null;
 };
 
 TLSSocket.prototype.getCipher = function(err) {
-  if (this.ssl) {
-    return this.ssl.getCurrentCipher();
+  if (this._handle) {
+    return this._handle.getCurrentCipher();
   } else {
     return null;
   }
 };
 
+TLSSocket.prototype.getEphemeralKeyInfo = function() {
+  if (this._handle)
+    return this._handle.getEphemeralKeyInfo();
+
+  return null;
+};
+
+TLSSocket.prototype.getProtocol = function() {
+  if (this._handle)
+    return this._handle.getProtocol();
+
+  return null;
+};
+
 // TODO: support anonymous (nocert) and PSK
 
 
@@ -585,10 +737,11 @@ TLSSocket.prototype.getCipher = function(err) {
 //
 function Server(/* [options], listener */) {
   var options, listener;
-  if (util.isObject(arguments[0])) {
+
+  if (arguments[0] !== null && typeof arguments[0] === 'object') {
     options = arguments[0];
     listener = arguments[1];
-  } else if (util.isFunction(arguments[0])) {
+  } else if (typeof arguments[0] === 'function') {
     options = {};
     listener = arguments[0];
   }
@@ -621,7 +774,7 @@ function Server(/* [options], listener */) {
 
   var timeout = options.handshakeTimeout || (120 * 1000);
 
-  if (!util.isNumber(timeout)) {
+  if (typeof timeout !== 'number') {
     throw new TypeError('handshakeTimeout must be a number');
   }
 
@@ -643,12 +796,13 @@ function Server(/* [options], listener */) {
       rejectUnauthorized: self.rejectUnauthorized,
       handshakeTimeout: timeout,
       NPNProtocols: self.NPNProtocols,
+      ALPNProtocols: self.ALPNProtocols,
       SNICallback: options.SNICallback || SNICallback
     });
 
     socket.on('secure', function() {
       if (socket._requestCert) {
-        var verifyError = socket.ssl.verifyError();
+        var verifyError = socket._handle.verifyError();
         if (verifyError) {
           socket.authorizationError = verifyError.code;
 
@@ -664,7 +818,11 @@ function Server(/* [options], listener */) {
     });
 
     var errorEmitted = false;
-    socket.on('close', function() {
+    socket.on('close', function(err) {
+      // Closed because of error - no need to emit it twice
+      if (err)
+        return;
+
       // Emit ECONNRESET
       if (!socket._controlReleased && !errorEmitted) {
         errorEmitted = true;
@@ -696,24 +854,34 @@ exports.createServer = function(options, listener) {
 
 Server.prototype._getServerData = function() {
   return {
-    ticketKeys: this._sharedCreds.context.getTicketKeys().toString('hex')
+    ticketKeys: this.getTicketKeys().toString('hex')
   };
 };
 
 
 Server.prototype._setServerData = function(data) {
-  this._sharedCreds.context.setTicketKeys(new Buffer(data.ticketKeys, 'hex'));
+  this.setTicketKeys(new Buffer(data.ticketKeys, 'hex'));
+};
+
+
+Server.prototype.getTicketKeys = function getTicketKeys(keys) {
+  return this._sharedCreds.context.getTicketKeys(keys);
+};
+
+
+Server.prototype.setTicketKeys = function setTicketKeys(keys) {
+  this._sharedCreds.context.setTicketKeys(keys);
 };
 
 
 Server.prototype.setOptions = function(options) {
-  if (util.isBoolean(options.requestCert)) {
+  if (typeof options.requestCert === 'boolean') {
     this.requestCert = options.requestCert;
   } else {
     this.requestCert = false;
   }
 
-  if (util.isBoolean(options.rejectUnauthorized)) {
+  if (typeof options.rejectUnauthorized === 'boolean') {
     this.rejectUnauthorized = options.rejectUnauthorized;
   } else {
     this.rejectUnauthorized = false;
@@ -727,33 +895,24 @@ Server.prototype.setOptions = function(options) {
   if (options.secureProtocol) this.secureProtocol = options.secureProtocol;
   if (options.crl) this.crl = options.crl;
   if (options.ciphers) this.ciphers = options.ciphers;
-  if (!util.isUndefined(options.ecdhCurve))
+  if (options.ecdhCurve !== undefined)
     this.ecdhCurve = options.ecdhCurve;
   if (options.dhparam) this.dhparam = options.dhparam;
   if (options.sessionTimeout) this.sessionTimeout = options.sessionTimeout;
   if (options.ticketKeys) this.ticketKeys = options.ticketKeys;
-
-  var secureOptions = common._getSecureOptions(options.secureProtocol,
-                                               options.secureOptions);
-
-  if (options.honorCipherOrder) {
-    secureOptions |= constants.SSL_OP_CIPHER_SERVER_PREFERENCE;
-  }
-
-  if (options.honorCipherOrder)
-    this.honorCipherOrder = true;
+  var secureOptions = options.secureOptions || 0;
+  if (options.honorCipherOrder !== undefined)
+    this.honorCipherOrder = !!options.honorCipherOrder;
   else
-    this.honorCipherOrder = false;
-
-  this.secureOptions = secureOptions;
-
+    this.honorCipherOrder = true;
+  if (secureOptions) this.secureOptions = secureOptions;
   if (options.NPNProtocols) tls.convertNPNProtocols(options.NPNProtocols, this);
+  if (options.ALPNProtocols)
+    tls.convertALPNProtocols(options.ALPNProtocols, this);
   if (options.sessionIdContext) {
     this.sessionIdContext = options.sessionIdContext;
   } else {
-    this.sessionIdContext = crypto.createHash('md5')
-                                  .update(process.argv.join(' '))
-                                  .digest('hex');
+    this.sessionIdContext = defaultSessionIdContext;
   }
 };
 
@@ -774,7 +933,7 @@ function SNICallback(servername, callback) {
   var ctx;
 
   this.server._contexts.some(function(elem) {
-    if (!util.isNull(servername.match(elem[0]))) {
+    if (servername.match(elem[0]) !== null) {
       ctx = elem[1];
       return true;
     }
@@ -803,107 +962,68 @@ function normalizeConnectArgs(listArgs) {
   var options = args[0];
   var cb = args[1];
 
-  if (util.isObject(listArgs[1])) {
+  if (listArgs[1] !== null && typeof listArgs[1] === 'object') {
     options = util._extend(options, listArgs[1]);
-  } else if (util.isObject(listArgs[2])) {
+  } else if (listArgs[2] !== null && typeof listArgs[2] === 'object') {
     options = util._extend(options, listArgs[2]);
   }
 
   return (cb) ? [options, cb] : [options];
 }
 
-function legacyConnect(hostname, options, NPN, context) {
-  assert(options.socket);
-  if (!tls_legacy)
-    tls_legacy = require('_tls_legacy');
-
-  var pair = tls_legacy.createSecurePair(context,
-                                         false,
-                                         true,
-                                         !!options.rejectUnauthorized,
-                                         {
-                                           NPNProtocols: NPN.NPNProtocols,
-                                           servername: hostname
-                                         });
-  tls_legacy.pipe(pair, options.socket);
-  pair.cleartext._controlReleased = true;
-  pair.on('error', function(err) {
-    pair.cleartext.emit('error', err);
-  });
-
-  return pair;
-}
-
 exports.connect = function(/* [port, host], options, cb */) {
-  var args = normalizeConnectArgs(arguments);
+  const argsLen = arguments.length;
+  var args = new Array(argsLen);
+  for (var i = 0; i < argsLen; i++)
+    args[i] = arguments[i];
+  args = normalizeConnectArgs(args);
   var options = args[0];
   var cb = args[1];
 
   var defaults = {
     rejectUnauthorized: '0' !== process.env.NODE_TLS_REJECT_UNAUTHORIZED,
     ciphers: tls.DEFAULT_CIPHERS,
-    checkServerIdentity: tls.checkServerIdentity
+    checkServerIdentity: tls.checkServerIdentity,
+    minDHSize: 1024
   };
 
   options = util._extend(defaults, options || {});
-
-  options.secureOptions = common._getSecureOptions(options.secureProtocol,
-                                                   options.secureOptions);
+  if (!options.keepAlive)
+    options.singleUse = true;
 
   assert(typeof options.checkServerIdentity === 'function');
+  assert(typeof options.minDHSize === 'number',
+         'options.minDHSize is not a number: ' + options.minDHSize);
+  assert(options.minDHSize > 0,
+         'options.minDHSize is not a positive number: ' +
+         options.minDHSize);
 
   var hostname = options.servername ||
                  options.host ||
-                 options.socket && options.socket._host,
-      NPN = {},
-      context = tls.createSecureContext(options);
+                 (options.socket && options.socket._host) ||
+                 'localhost';
+  const NPN = {};
+  const ALPN = {};
+  const context = options.secureContext || tls.createSecureContext(options);
   tls.convertNPNProtocols(options.NPNProtocols, NPN);
+  tls.convertALPNProtocols(options.ALPNProtocols, ALPN);
 
-  // Wrapping TLS socket inside another TLS socket was requested -
-  // create legacy secure pair
-  var socket;
-  var legacy;
-  var result;
-  if (options.socket instanceof TLSSocket) {
-    debug('legacy connect');
-    legacy = true;
-    socket = legacyConnect(hostname, options, NPN, context);
-    result = socket.cleartext;
-  } else {
-    legacy = false;
-    socket = new TLSSocket(options.socket, {
-      secureContext: context,
-      isServer: false,
-      requestCert: true,
-      rejectUnauthorized: options.rejectUnauthorized,
-      session: options.session,
-      NPNProtocols: NPN.NPNProtocols,
-      requestOCSP: options.requestOCSP
-    });
-    result = socket;
-  }
-
-  if (socket._handle && !socket._connecting) {
-    onHandle();
-  } else {
-    // Not even started connecting yet (or probably resolving dns address),
-    // catch socket errors and assign handle.
-    if (!legacy && options.socket) {
-      options.socket.once('connect', function() {
-        assert(options.socket._handle);
-        socket._handle = options.socket._handle;
-        socket._handle.owner = socket;
-        socket.emit('connect');
-      });
-    }
-    socket.once('connect', onHandle);
-  }
+  var socket = new TLSSocket(options.socket, {
+    pipe: options.path && !options.port,
+    secureContext: context,
+    isServer: false,
+    requestCert: true,
+    rejectUnauthorized: options.rejectUnauthorized,
+    session: options.session,
+    NPNProtocols: NPN.NPNProtocols,
+    ALPNProtocols: ALPN.ALPNProtocols,
+    requestOCSP: options.requestOCSP
+  });
 
   if (cb)
-    result.once('secureConnect', cb);
+    socket.once('secureConnect', cb);
 
   if (!options.socket) {
-    assert(!legacy);
     var connect_opt;
     if (options.path && !options.port) {
       connect_opt = { path: options.path };
@@ -914,63 +1034,72 @@ exports.connect = function(/* [port, host], options, cb */) {
         localAddress: options.localAddress
       };
     }
-    socket.connect(connect_opt);
-  }
-
-  return result;
-
-  function onHandle() {
-    if (!legacy)
-      socket._releaseControl();
-
-    if (options.session)
-      socket.setSession(options.session);
-
-    if (!legacy) {
-      if (options.servername)
-        socket.setServername(options.servername);
-
+    socket.connect(connect_opt, function() {
       socket._start();
-    }
-    socket.on('secure', function() {
-      var verifyError = socket.ssl.verifyError();
-
-      // Verify that server's identity matches it's certificate's names
-      if (!verifyError) {
-        var cert = result.getPeerCertificate();
-        verifyError = options.checkServerIdentity(hostname, cert);
-      }
-
-      if (verifyError) {
-        result.authorized = false;
-        result.authorizationError = verifyError.code || verifyError.message;
-
-        if (options.rejectUnauthorized) {
-          result.emit('error', verifyError);
-          result.destroy();
-          return;
-        } else {
-          result.emit('secureConnect');
-        }
-      } else {
-        result.authorized = true;
-        result.emit('secureConnect');
-      }
-
-      // Uncork incoming data
-      result.removeListener('end', onHangUp);
     });
-
-    function onHangUp() {
-      // NOTE: This logic is shared with _http_client.js
-      if (!socket._hadError) {
-        socket._hadError = true;
-        var error = new Error('socket hang up');
-        error.code = 'ECONNRESET';
-        socket.destroy();
-        socket.emit('error', error);
-      }
-    }
-    result.once('end', onHangUp);
   }
+
+  socket._releaseControl();
+
+  if (options.session)
+    socket.setSession(options.session);
+
+  if (options.servername)
+    socket.setServername(options.servername);
+
+  if (options.socket)
+    socket._start();
+
+  socket.on('secure', function() {
+    // Check the size of DHE parameter above minimum requirement
+    // specified in options.
+    var ekeyinfo = socket.getEphemeralKeyInfo();
+    if (ekeyinfo.type === 'DH' && ekeyinfo.size < options.minDHSize) {
+      var err = new Error('DH parameter size ' + ekeyinfo.size +
+                          ' is less than ' + options.minDHSize);
+      socket.emit('error', err);
+      socket.destroy();
+      return;
+    }
+
+    var verifyError = socket._handle.verifyError();
+
+    // Verify that server's identity matches it's certificate's names
+    // Unless server has resumed our existing session
+    if (!verifyError && !socket.isSessionReused()) {
+      var cert = socket.getPeerCertificate();
+      verifyError = options.checkServerIdentity(hostname, cert);
+    }
+
+    if (verifyError) {
+      socket.authorized = false;
+      socket.authorizationError = verifyError.code || verifyError.message;
+
+      if (options.rejectUnauthorized) {
+        socket.destroy(verifyError);
+        return;
+      } else {
+        socket.emit('secureConnect');
+      }
+    } else {
+      socket.authorized = true;
+      socket.emit('secureConnect');
+    }
+
+    // Uncork incoming data
+    socket.removeListener('end', onHangUp);
+  });
+
+  function onHangUp() {
+    // NOTE: This logic is shared with _http_client.js
+    if (!socket._hadError) {
+      socket._hadError = true;
+      var error = new Error('socket hang up');
+      error.code = 'ECONNRESET';
+      socket.destroy(error);
+    }
+  }
+  socket.once('end', onHangUp);
+
+  return socket;
 };
