@@ -141,7 +141,7 @@ QueryString.escape = function(str) {
     if (i < str.length)
       c2 = str.charCodeAt(i) & 0x3FF;
     else
-      c2 = 0;
+      throw new URIError('URI malformed');
     lastPos = i + 1;
     c = 0x10000 + (((c & 0x3FF) << 10) | c2);
     out += hexTable[0xF0 | (c >> 18)] +
