@@ -75,7 +75,7 @@ exports.checkServerIdentity = function checkServerIdentity(host, cert) {
   // Create regexp to much hostnames
   function regexpify(host, wildcards) {
     // Add trailing dot (make hostnames uniform)
-    if (!/\.$/.test(host)) host += '.';
+    if (!host || !host.endsWith('.')) host += '.';
 
     // The same applies to hostname with more than one wildcard,
     // if hostname has wildcard when wildcards are not allowed,
@@ -144,7 +144,7 @@ exports.checkServerIdentity = function checkServerIdentity(host, cert) {
     }
   } else if (cert.subject) {
     // Transform hostname to canonical form
-    if (!/\.$/.test(host)) host += '.';
+    if (!host || !host.endsWith('.')) host += '.';
 
     // Otherwise check all DNS/URI records from certificate
     // (with allowed wildcards)
