@@ -5,41 +5,77 @@ const Buffer = require('buffer').Buffer;
 const internalUtil = require('internal/util');
 const binding = process.binding('util');
 
+const isError = internalUtil.isError;
+
 var Debug;
 
-const formatRegExp = /%[sdj%]/g;
+function tryStringify(arg) {
+  try {
+    return JSON.stringify(arg);
+  } catch (_) {
+    return '[Circular]';
+  }
+}
+
 exports.format = function(f) {
   if (typeof f !== 'string') {
-    var objects = [];
-    for (var i = 0; i < arguments.length; i++) {
-      objects.push(inspect(arguments[i]));
+    const objects = new Array(arguments.length);
+    for (var index = 0; index < arguments.length; index++) {
+      objects[index] = inspect(arguments[index]);
     }
     return objects.join(' ');
   }
 
-  if (arguments.length === 1) return f;
+  var argLen = arguments.length;
 
-  var i = 1;
-  var args = arguments;
-  var len = args.length;
-  var str = String(f).replace(formatRegExp, function(x) {
-    if (x === '%%') return '%';
-    if (i >= len) return x;
-    switch (x) {
-      case '%s': return String(args[i++]);
-      case '%d': return Number(args[i++]);
-      case '%j':
-        try {
-          return JSON.stringify(args[i++]);
-        } catch (_) {
-          return '[Circular]';
-        }
-        // falls through
-      default:
-        return x;
+  if (argLen === 1) return f;
+
+  var str = '';
+  var a = 1;
+  var lastPos = 0;
+  for (var i = 0; i < f.length;) {
+    if (f.charCodeAt(i) === 37/*'%'*/ && i + 1 < f.length) {
+      switch (f.charCodeAt(i + 1)) {
+        case 100: // 'd'
+          if (a >= argLen)
+            break;
+          if (lastPos < i)
+            str += f.slice(lastPos, i);
+          str += Number(arguments[a++]);
+          lastPos = i = i + 2;
+          continue;
+        case 106: // 'j'
+          if (a >= argLen)
+            break;
+          if (lastPos < i)
+            str += f.slice(lastPos, i);
+          str += tryStringify(arguments[a++]);
+          lastPos = i = i + 2;
+          continue;
+        case 115: // 's'
+          if (a >= argLen)
+            break;
+          if (lastPos < i)
+            str += f.slice(lastPos, i);
+          str += String(arguments[a++]);
+          lastPos = i = i + 2;
+          continue;
+        case 37: // '%'
+          if (lastPos < i)
+            str += f.slice(lastPos, i);
+          str += '%';
+          lastPos = i = i + 2;
+          continue;
+      }
     }
-  });
-  for (var x = args[i]; i < len; x = args[++i]) {
+    ++i;
+  }
+  if (lastPos === 0)
+    str = f;
+  else if (lastPos < f.length)
+    str += f.slice(lastPos);
+  while (a < argLen) {
+    const x = arguments[a++];
     if (x === null || (typeof x !== 'object' && typeof x !== 'symbol')) {
       str += ' ' + x;
     } else {
@@ -75,7 +111,7 @@ exports.debuglog = function(set) {
 
 
 /**
- * Echos the value of a value. Trys to print the value out
+ * Echos the value of a value. Tries to print the value out
  * in the best way possible given the different types.
  *
  * @param {Object} obj The object to print out.
@@ -159,11 +195,12 @@ function stylizeNoColor(str, styleType) {
 
 
 function arrayToHash(array) {
-  var hash = {};
+  var hash = Object.create(null);
 
-  array.forEach(function(val, idx) {
+  for (var i = 0; i < array.length; i++) {
+    var val = array[i];
     hash[val] = true;
-  });
+  }
 
   return hash;
 }
@@ -195,6 +232,7 @@ function ensureDebugIsInitialized() {
 
 function inspectPromise(p) {
   ensureDebugIsInitialized();
+  // Only create a mirror if the object is a Promise.
   if (!binding.isPromise(p))
     return null;
   const mirror = Debug.MakeMirror(p, true);
@@ -250,7 +288,7 @@ function formatValue(ctx, value, recurseTimes) {
 
   if (typeof raw === 'string') {
     // for boxed Strings, we have to remove the 0-n indexed entries,
-    // since they just noisey up the output and are redundant
+    // since they just noisy up the output and are redundant
     keys = keys.filter(function(key) {
       return !(key >= 0 && key < raw.length);
     });
@@ -289,16 +327,19 @@ function formatValue(ctx, value, recurseTimes) {
   var constructor = getConstructorOf(value);
   var base = '', empty = false, braces, formatter;
 
+  // We can't compare constructors for various objects using a comparison like
+  // `constructor === Array` because the object could have come from a different
+  // context and thus the constructor won't match. Instead we check the
+  // constructor names (including those up the prototype chain where needed) to
+  // determine object types.
   if (Array.isArray(value)) {
-    // We can't use `constructor === Array` because this could
-    // have come from a Debug context.
-    // Otherwise, an Array will print "Array [...]".
+    // Unset the constructor to prevent "Array [...]" for ordinary arrays.
     if (constructor && constructor.name === 'Array')
       constructor = null;
     braces = ['[', ']'];
     empty = value.length === 0;
     formatter = formatArray;
-  } else if (value instanceof Set) {
+  } else if (binding.isSet(value)) {
     braces = ['{', '}'];
     // With `showHidden`, `length` will display as a hidden property for
     // arrays. For consistency's sake, do the same for `size`, even though this
@@ -307,7 +348,7 @@ function formatValue(ctx, value, recurseTimes) {
       keys.unshift('size');
     empty = value.size === 0;
     formatter = formatSet;
-  } else if (value instanceof Map) {
+  } else if (binding.isMap(value)) {
     braces = ['{', '}'];
     // Ditto.
     if (ctx.showHidden)
@@ -315,8 +356,7 @@ function formatValue(ctx, value, recurseTimes) {
     empty = value.size === 0;
     formatter = formatMap;
   } else {
-    // Only create a mirror if the object superficially looks like a Promise.
-    var promiseInternals = value instanceof Promise && inspectPromise(value);
+    var promiseInternals = inspectPromise(value);
     if (promiseInternals) {
       braces = ['{', '}'];
       formatter = formatPromise;
@@ -332,7 +372,8 @@ function formatValue(ctx, value, recurseTimes) {
         empty = false;
         formatter = formatCollectionIterator;
       } else {
-        if (constructor === Object)
+        // Unset the constructor to prevent "Object {...}" for ordinary objects.
+        if (constructor && constructor.name === 'Object')
           constructor = null;
         braces = ['{', '}'];
         empty = true;  // No other data than keys.
@@ -517,7 +558,7 @@ function formatCollectionIterator(ctx, value, recurseTimes, visibleKeys, keys) {
   var nextRecurseTimes = recurseTimes === null ? null : recurseTimes - 1;
   var vals = mirror.preview();
   var output = [];
-  for (let o of vals) {
+  for (const o of vals) {
     output.push(formatValue(ctx, o, nextRecurseTimes));
   }
   return output;
@@ -666,7 +707,7 @@ function isUndefined(arg) {
 exports.isUndefined = isUndefined;
 
 function isRegExp(re) {
-  return objectToString(re) === '[object RegExp]';
+  return binding.isRegExp(re);
 }
 exports.isRegExp = isRegExp;
 
@@ -676,13 +717,10 @@ function isObject(arg) {
 exports.isObject = isObject;
 
 function isDate(d) {
-  return objectToString(d) === '[object Date]';
+  return binding.isDate(d);
 }
 exports.isDate = isDate;
 
-function isError(e) {
-  return objectToString(e) === '[object Error]' || e instanceof Error;
-}
 exports.isError = isError;
 
 function isFunction(arg) {
@@ -698,10 +736,6 @@ exports.isPrimitive = isPrimitive;
 
 exports.isBuffer = Buffer.isBuffer;
 
-function objectToString(o) {
-  return Object.prototype.toString.call(o);
-}
-
 
 function pad(n) {
   return n < 10 ? '0' + n.toString(10) : n.toString(10);
@@ -757,14 +791,7 @@ exports.inherits = function(ctor, superCtor) {
                         'have a prototype.');
 
   ctor.super_ = superCtor;
-  ctor.prototype = Object.create(superCtor.prototype, {
-    constructor: {
-      value: ctor,
-      enumerable: false,
-      writable: true,
-      configurable: true
-    }
-  });
+  Object.setPrototypeOf(ctor.prototype, superCtor.prototype);
 };
 
 exports._extend = function(origin, add) {
@@ -786,13 +813,6 @@ function hasOwnProperty(obj, prop) {
 
 // Deprecated old stuff.
 
-exports.p = internalUtil.deprecate(function() {
-  for (var i = 0, len = arguments.length; i < len; ++i) {
-    console.error(exports.inspect(arguments[i]));
-  }
-}, 'util.p is deprecated. Use console.error instead.');
-
-
 exports.exec = internalUtil.deprecate(function() {
   return require('child_process').exec.apply(this, arguments);
 }, 'util.exec is deprecated. Use child_process.exec instead.');
