@@ -41,6 +41,9 @@ enum node_zlib_mode {
   UNZIP
 };
 
+#define GZIP_HEADER_ID1 0x1f
+#define GZIP_HEADER_ID2 0x8b
+#define GZIP_MIN_HEADER_SIZE 10
 
 void InitZlib(v8::Local<v8::Object> target);
 
@@ -254,6 +257,20 @@ class ZCtx : public AsyncWrap {
             ctx->err_ = Z_NEED_DICT;
           }
         }
+        while (ctx->strm_.avail_in >= GZIP_MIN_HEADER_SIZE &&
+               ctx->mode_ == GUNZIP &&
+               ctx->err_ == Z_STREAM_END) {
+          // Bytes remain in input buffer. Perhaps this is another compressed
+          // member in the same archive, or just trailing garbage.
+          // Check the header to find out.
+          if (ctx->strm_.next_in[0] != GZIP_HEADER_ID1 ||
+              ctx->strm_.next_in[1] != GZIP_HEADER_ID2) {
+            // Not a valid gzip member
+            break;
+          }
+          Reset(ctx);
+          ctx->err_ = inflate(&ctx->strm_, ctx->flush_);
+        }
         break;
       default:
         CHECK(0 && "wtf?");
@@ -524,10 +541,12 @@ class ZCtx : public AsyncWrap {
     switch (ctx->mode_) {
       case DEFLATE:
       case DEFLATERAW:
+      case GZIP:
         ctx->err_ = deflateReset(&ctx->strm_);
         break;
       case INFLATE:
       case INFLATERAW:
+      case GUNZIP:
         ctx->err_ = inflateReset(&ctx->strm_);
         break;
       default:
